from argparse import ArgumentParser, _SubParsersAction


def get_mdmix_parser():
    parser = ArgumentParser("MDMix")
    parser.add_argument(
        "-c", "--config",
        help=(
            "location of additional configuration file. "
            "if no config file is provided, all the config values will be the default ones. "
            "default is config.yml"
        ),
        default=None
    )
    parser.add_argument(
        "-v", "--verbosity",
        help="set verbosity level. default is INFO",
        default="INFO",
        choices=["INFO", "WARNING", "ERROR", "CRITICAL"]
    )
    parser.add_argument(
        "-l", "--log",
        help="send the logging output to the specified file",
        default=None
    )

    return parser


MDMIX_PARSER = get_mdmix_parser()


def get_plugin_subparsers(parser: ArgumentParser = None):
    parser = parser if parser is not None else MDMIX_PARSER
    actions = parser._subparsers._actions if parser._subparsers is not None else []
    subparser = next(
        iter(
            [action for action in actions if isinstance(action, _SubParsersAction)],
        ),
        None
    )
    if subparser is None:
        subparser = parser.add_subparsers(title="plugin", dest="plugin")
    return subparser



from typing import List, Optional
from argparse import ArgumentParser, Namespace, _SubParsersAction
import logging
import sys
import yaml
from pymdmix_core.plugin.base import Plugin, PluginAction, PluginManager, MDMIX_PLUGIN_MANAGER
from pymdmix_core.settings import SETTINGS
from pymdmix_core.parser import MDMIX_PARSER

logger = logging.getLogger(__name__)


class MDMix:

    def __init__(self, config: Optional[str] = None) -> None:
        if config is not None:
            SETTINGS.update_settings_with_file(config)

        self.plugin_manager = MDMIX_PLUGIN_MANAGER
        for plugin in SETTINGS["pymdmix_core"]["installed_plugins"]:
            logger.info(f"loading plugin: {plugin}")
            self.plugin_manager.load_plugin(plugin)

    def run(self) -> None:
        args = MDMIX_PARSER.parse_args()
        plugin = self.plugin_manager.plugins.get(args.plugin)
        if plugin is None:
            MDMIX_PARSER.print_help(sys.stderr)
            return

        plugin.run(args)


class ActionList(PluginAction):

    ACTION_NAME = "list"

    def run(self, args: Namespace) -> None:
        print("Available plugins:")
        for plugin_name in MDMIX_PLUGIN_MANAGER.plugins:
            print(f"\t- {plugin_name}")


class ActionLoad(PluginAction):

    ACTION_NAME = "load"

    def __init__(self, subparser: _SubParsersAction, plugin_manager: PluginManager) -> None:
        super().__init__(subparser)
        self.plugin_manager = plugin_manager

    def init_parser(self):
        self.parser.add_argument("plugin_module")

    def run(self, args: Namespace) -> None:
        try:
            self.plugin_manager.load_plugin(args.plugin_module)
        except Exception as e:
            logger.error(f"Unable to load plugin module {args.plugin_module}: {repr(e)}")
            raise e


class ActionAdd(ActionLoad):

    ACTION_NAME = "add"

    def run(self, args: Namespace) -> None:
        try:
            super().run(args)
            defaults_file = SETTINGS.defaults_filename
            defaults = {}
            with open(defaults_file, 'r') as input_defaults_file:
                defaults = yaml.full_load(input_defaults_file)
            installed = defaults["pymdmix_core"]["installed_plugins"]
            if args.plugin_module not in installed:
                installed.append(args.plugin_module)
            with open(defaults_file, 'w') as output_defaults_file:
                yaml.dump(defaults, output_defaults_file)
        except Exception as e:
            logger.error(f"Unable to add plugin module {args.plugin_module} to installed plugins: {repr(e)}")
            raise e


class ActionRemove(ActionLoad):

    ACTION_NAME = "remove"

    def run(self, args: Namespace) -> None:
        try:
            if args.plugin_module is None:
                self.parser.print_help(sys.stderr)
                raise Exception("no plugin module provided")
            defaults_file = SETTINGS.defaults_filename
            defaults = {}
            with open(defaults_file, 'r') as input_defaults_file:
                defaults = yaml.full_load(input_defaults_file)
            installed: List = defaults["pymdmix_core"]["installed_plugins"]
            if args.plugin_module not in installed:
                raise Exception("Not found")
            installed.remove(args.plugin_module)
            with open(defaults_file, 'w') as output_defaults_file:
                yaml.dump(defaults, output_defaults_file)
        except Exception as e:
            logger.error(f"Unable to remove plugin module {args.plugin_module}: {repr(e)}")
            raise e


class CorePlugin(Plugin):

    NAME = "plugin"

    def __init__(self, master_parser: ArgumentParser, plugin_manager: Optional[PluginManager] = None) -> None:
        self.plugin_manager = plugin_manager if plugin_manager is not None else MDMIX_PLUGIN_MANAGER
        super().__init__(master_parser)

    def init_actions(self, action_subparser):
        self.register_action(ActionList(action_subparser))
        self.register_action(ActionLoad(action_subparser, self.plugin_manager))
        self.register_action(ActionAdd(action_subparser, self.plugin_manager))
        self.register_action(ActionRemove(action_subparser, self.plugin_manager))



import sqlalchemy

from pymdmix_core.settings import SETTINGS

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker


def get_sql_uri() -> str:
    dialect = SETTINGS["pymdmix_core"]["orm"]["dialect"]
    if dialect == "sqlite":
        db_filename = SETTINGS.get_file(SETTINGS["pymdmix_core"]["orm"]["db_file"])
        return f"sqlite:///{db_filename}"
    else:
        raise NotImplementedError(f"the dialect {dialect} for sql orm is not supported yet")


SQL_ENGINE = sqlalchemy.create_engine(get_sql_uri())
SQL_SESSION = sessionmaker(bind=SQL_ENGINE)()

BaseModel = declarative_base()



import logging
import yaml
import json
from argparse import Namespace, ArgumentParser
from typing import Dict, Optional, Type
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker
from pymdmix_core.plugin.base import PluginAction, _SubParsersAction
from pymdmix_core.orm import BaseModel, SQL_ENGINE

from pymdmix_core.plugin import Plugin

logger = logging.getLogger(__name__)


def parse_file_from_args(args: Namespace) -> Optional[Dict]:
    if args.yaml is not None:
        with open(args.yaml, 'r') as yml_file:
            return yaml.full_load(yml_file)
    elif args.json is not None:
        with open(args.json, 'r') as json_file:
            return json.load(json_file)


class ActionCreate(PluginAction):

    ACTION_NAME = "create"

    def __init__(self, subparser: _SubParsersAction, parent_plugin: 'CRUDPlugin') -> None:
        super().__init__(subparser)
        self.parent_plugin = parent_plugin

    def run(self, args: Namespace) -> None:
        model = self.parent_plugin.factory(args)
        self.parent_plugin.session.commit()
        if model is None:
            logger.critical("Could not create model")
            exit(1)

    def init_parser(self):
        self.parser.add_argument(
            "--yaml", "-y",
            help="A yaml file with all the necessary parameters to create the new item"
        )
        self.parser.add_argument(
            "--json", "-j",
            help="A json file with all the necessary parameters to create the new item"
        )


class ActionRead(PluginAction):

    ACTION_NAME = "info"

    def __init__(self, subparser: _SubParsersAction, parent_plugin: 'CRUDPlugin') -> None:
        super().__init__(subparser)
        self.parent_plugin = parent_plugin

    def init_parser(self):
        super().init_parser()
        self.parser.add_argument("id", nargs="+")

    def run(self, args: Namespace) -> None:
        session = self.parent_plugin.session
        model_class = self.parent_plugin.CLASS
        models = session.query(model_class).filter(model_class.id.in_(args.id)).all()
        for model in models:
            print(model.__repr__())


# We will disallow this for now.
#
# class ActionUpdate(PluginAction):
#
#     ACTION_NAME = "update"
#
#     def __init__(self, subparser: _SubParsersAction, parent_plugin: 'CRUDPlugin') -> None:
#         super().__init__(subparser)
#         self.parent_plugin = parent_plugin


class ActionDelete(PluginAction):

    ACTION_NAME = "delete"

    def __init__(self, subparser: _SubParsersAction, parent_plugin: 'CRUDPlugin') -> None:
        super().__init__(subparser)
        self.parent_plugin = parent_plugin

    def init_parser(self):
        super().init_parser()
        self.parser.add_argument("id", nargs="+")

    def run(self, args: Namespace) -> None:
        session = self.parent_plugin.session
        model_class = self.parent_plugin.CLASS
        query = session.query(model_class).filter(model_class.id.in_(args.id))
        query.delete(synchronize_session='fetch')
        session.commit()


class ActionList(PluginAction):

    ACTION_NAME = "list"

    def __init__(self, subparser: _SubParsersAction, parent_plugin: 'CRUDPlugin') -> None:
        super().__init__(subparser)
        self.parent_plugin = parent_plugin

    def run(self, _: Namespace) -> None:
        session = self.parent_plugin.session
        model_class = self.parent_plugin.CLASS
        for model in session.query(model_class).all():
            print(model)


class CRUDPlugin(Plugin):

    NAME = "crud_plugin"
    CLASS: Type[BaseModel] = BaseModel

    def factory(self, args: Namespace) -> Optional[BaseModel]:
        """
        A default factory. If a yaml or json file is passed as argument
        It will be parsed into a dictionary and its key-value pairs passed
        as arguments to the CLASS constructor.
        Override this method for a more refined behavior
        """
        fields = parse_file_from_args(args)
        if fields is not None:
            model = self.CLASS(**fields)
            self.session.add(model)
            self.session.commit()
            return model

    def __init__(self, master_parser: ArgumentParser, engine: Optional[Engine] = None) -> None:
        super().__init__(master_parser)
        self.engine = engine if engine is not None else SQL_ENGINE
        self.session = sessionmaker(bind=self.engine)()
        self.CLASS.metadata.create_all(bind=self.engine)

    def init_actions(self, action_subparser):
        self.register_action(ActionCreate(action_subparser, self))
        self.register_action(ActionRead(action_subparser, self))
        # disabled for now
        # self.register_action(ActionUpdate(action_subparser, self))
        self.register_action(ActionDelete(action_subparser, self))
        self.register_action(ActionList(action_subparser, self))



from typing import Dict, List, Optional, Type
import logging
import sys
from argparse import ArgumentParser, Namespace, _SubParsersAction
from abc import abstractmethod
from importlib import import_module

from pymdmix_core.settings import SETTINGS
from pymdmix_core.parser import MDMIX_PARSER, get_plugin_subparsers


logger = logging.getLogger(__name__)


class PluginAction:

    ACTION_NAME: str = "action"

    def __init__(self, subparser: _SubParsersAction) -> None:
        self.parser = subparser.add_parser(self.ACTION_NAME, parents=self.get_parents())
        self.init_parser()

    @abstractmethod
    def run(self, args: Namespace) -> None:
        pass

    def init_parser(self):
        pass

    def get_parents(self) -> List[ArgumentParser]:
        return []


class Plugin:

    NAME: str = "plugin"
    HELP_STRING: str = "plugin help"
    LOAD_CONFIG: bool = False
    CONFIG_FILE: str = "pymdmix_core.yml"
    ALLOW_EMPTY_ACTION = False

    def init_parser(self) -> None:
        """
        override this method to configure options for the plugin parser other than actions parsers.
        plugin parser passed as parameter.
        """
        pass

    def init_actions(self, action_subparser):
        """
        Override this method to initialize and register actions associated to this plugin.
        This function will be called on super().__init__
        """
        pass

    def __init__(self, master_parser: ArgumentParser) -> None:
        self.load_config()
        self.actions: Dict[str, PluginAction] = {}
        self.master_parser = master_parser
        subparser = get_plugin_subparsers(self.master_parser)
        self.parser = subparser.add_parser(self.NAME)
        self.init_parser()
        self.action_subparser = self.parser.add_subparsers(dest="action", required=(not self.ALLOW_EMPTY_ACTION))
        self.init_actions(self.action_subparser)

    def register_action(self, action: PluginAction):
        self.actions[action.ACTION_NAME] = action

    def run(self, args: Namespace) -> None:
        action = self.actions.get(args.action)
        if action is None:
            if not self.ALLOW_EMPTY_ACTION:
                self.parser.print_help(sys.stderr)
                raise ValueError("Action is mandatory")
        else:
            action.run(args)

    def load_config(self) -> None:
        if self.LOAD_CONFIG:
            SETTINGS.update_settings_with_file(SETTINGS.get_defaults_filename(self.CONFIG_FILE))


class PluginManager:

    def __init__(self, parser: Optional[ArgumentParser] = None) -> None:
        self.plugins: Dict[str, Plugin] = {}
        self.parser = parser if parser is not None else MDMIX_PARSER

    def load_plugin(self, plugin_name: str):
        mod = import_module(plugin_name)
        plugin_class: Type[Plugin] = mod.get_plugin_class()
        self.plugins[plugin_class.NAME] = plugin_class(self.parser)


MDMIX_PLUGIN_MANAGER = PluginManager()



import os
import yaml
from typing import List, Optional
import sys
import site
import logging
from collections.abc import Mapping


logger = logging.getLogger(__name__)


class Settings:
    def __init__(self, filename: Optional[str] = None) -> None:
        self.defaults_filename = self.get_defaults_filename(filename)
        self.files = [self.defaults_filename]
        self.data: dict = self.load_data(self.defaults_filename)
        self.get = self.data.get
        self.home = os.path.dirname(self.defaults_filename)

    def __getitem__(self, k):
        return self.data[k]

    def __len__(self):
        return len(self.data)

    def __contains__(self, k):
        return k in self.data

    def update_settings_with_file(self, filename: str):
        self.merge(self.load_data(filename))
        self.files.append(filename)

    def get_file(self, filename: str):
        return os.path.join(self.home, filename)

    @staticmethod
    def get_defaults_filename(filename: Optional[str] = None) -> str:
        filename = filename if filename is not None else "pymdmix_core.yml"
        home_candidates = Settings.get_mdmix_homes()
        logger.info(f"Searching for mdmix home in: {home_candidates}")
        configs = [filename]
        if filename != os.path.abspath(filename):
            configs += [os.path.join(path, filename) for path in home_candidates]
        configs = [config for config in configs if os.path.exists(config)]
        if len(configs) > 0:
            logger.info(f"found configs: {configs}")
            config = configs[0]
            logger.info(f"Using {config} for default settings")
            return config
        else:
            logger.warning(f"no config file found in default locations for {filename}. trying current folder")
            return filename

    @staticmethod
    def get_mdmix_homes() -> List[str]:
        mdmix_home = os.getenv("MDMIX_HOME")
        paths = [sys.prefix, os.path.join(sys.prefix, "local")]
        paths += site.PREFIXES
        paths += ["/usr", "/usr/local"]
        paths = [os.path.join(path, "pymdmix") for path in paths]
        if mdmix_home is not None:
            paths = [mdmix_home] + paths
        return [path for path in paths if os.path.exists(path)]

    @staticmethod
    def load_data(filename: str) -> dict:
        with open(filename, "r") as file:
            d = yaml.load(file, Loader=yaml.FullLoader)
        return d if d is not None else {}

    def merge(self, data):
        def _merge(dict1, dict2):
            for key, value in dict2.items():
                if (
                    key in dict1
                    and isinstance(dict1[key], dict)
                    and isinstance(value, Mapping)
                ):
                    _merge(dict1[key], value)
                elif (
                    key in dict1
                    and isinstance(dict1[key], list)
                    and isinstance(value, list)
                ):
                    dict1[key] += value
                else:
                    dict1[key] = dict2[key]

        _merge(self.data, data)


SETTINGS = Settings()



#!/usr/bin/env python3

from setuptools import setup


def getRequirements():
    requirements = []
    with open("requirements.txt", "r") as reqfile:
        for line in reqfile.readlines():
            requirements.append(line.strip())
    return requirements


def getVersion():
    return "4.0.4"


setup(
    python_requires=">=3.8",
    name="pymdmix-core",
    zip_safe=False,
    version=getVersion(),
    description="Molecular Dynamics with organic solvent mixtures setup and analysis",
    author="ggutierrez-bio",
    author_email="",
    url="https://github.com/ggutierrez-bio/mdmix4",
    packages=["pymdmix_core", "pymdmix_core.plugin"],
    inlcude_package_data=True,
    data_files=[("pymdmix", ["defaults/pymdmix_core.yml"])],
    scripts=["bin/mdmix"],
    install_requires=getRequirements(),
)
